from __future__ import annotations

import datetime
import json as sysjson
import platform
import re
import socket
import subprocess
import sys
import typing
import uuid
from http.client import HTTPConnection, HTTPResponse
from typing import List, Optional, TypeVar, Union

from pydantic import BaseModel
from pygments import formatters, highlight, lexers

from uiautodev.exceptions import RequestError
from uiautodev.model import Node


def is_output_terminal() -> bool:
    """
    Check if the standard output is attached to a terminal.
    """
    return sys.stdout.isatty()


def enable_windows_ansi_support():
    if platform.system().lower() == "windows" and is_output_terminal():
        import ctypes

        kernel32 = ctypes.windll.kernel32
        kernel32.SetConsoleMode(kernel32.GetStdHandle(-11), 7)


def default_json_encoder(obj):
    if isinstance(obj, bytes):
        return f'<{obj.hex()}>'
    if isinstance(obj, datetime.datetime):
        return str(obj)
    if isinstance(obj, uuid.UUID):
        return str(obj)
    if isinstance(obj, BaseModel):
        return obj.model_dump()
    raise TypeError()


def print_json(buf, colored=None, default=default_json_encoder):
    """ copy from pymobiledevice3 """
    formatted_json = sysjson.dumps(buf, sort_keys=True, indent=4, default=default)
    if colored is None:
        if is_output_terminal():
            colored = True
            enable_windows_ansi_support()
        else:
            colored = False

    if colored:
        colorful_json = highlight(formatted_json, lexers.JsonLexer(),
                                  formatters.TerminalTrueColorFormatter(style='stata-dark'))
        print(colorful_json)
    else:
        print(formatted_json)


_T = TypeVar("_T")


def convert_to_type(value: str, _type: _T) -> _T:
    """ usage example:
    convert_to_type("123", int)
    """
    if _type in (int, float, str):
        return _type(value)
    if _type == bool:
        return value.lower() in ("true", "1")
    if _type == Union[int, float]:
        return float(value) if "." in value else int(value)
    if _type == re.Pattern:
        return re.compile(value)
    raise NotImplementedError(f"convert {value} to {_type}")


def convert_params_to_model(params: List[str], model: BaseModel) -> BaseModel:
    """ used in cli.py """
    assert len(params) > 0
    if len(params) == 1:
        try:
            return model.model_validate_json(params)
        except Exception as e:
            print("module_parse_error", e)

    value = {}
    type_hints = typing.get_type_hints(model)
    for p in params:
        if "=" not in p:
            _type = type_hints.get(p)
            if _type == bool:
                value[p] = True
                continue
            elif _type is None:
                print(f"unknown key: {p}")
                continue
            raise ValueError(f"missing value for {p}")
        k, v = p.split("=", 1)
        _type = type_hints.get(k)
        if _type is None:
            print(f"unknown key: {k}")
            continue
        value[k] = convert_to_type(v, _type)
    return model.model_validate(value)


class SocketHTTPConnection(HTTPConnection):
    def __init__(self, conn: socket.socket, timeout: float):
        super().__init__("localhost", timeout=timeout)
        self.__conn = conn

    def connect(self):
        self.sock = self.__conn

    def __enter__(self) -> HTTPConnection:
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class MySocketHTTPConnection(SocketHTTPConnection):
    def connect(self):
        super().connect()
        self.sock.settimeout(self.timeout)


def fetch_through_socket(sock: socket.socket, path: str, method: str = "GET", json: Optional[dict] = None,
                         timeout: float = 60) -> bytearray:
    """ usage example:
    with socket.create_connection((host, port)) as s:
        request_through_socket(s, "GET", "/")
    """
    conn = MySocketHTTPConnection(sock, timeout)
    try:
        if json is None:
            conn.request(method, path)
        else:
            conn.request(method, path, body=sysjson.dumps(json), headers={"Content-Type": "application/json"})
        response = conn.getresponse()
        if response.getcode() != 200:
            raise RequestError(f"request {method} {path}, status: {response.getcode()}")
        content = bytearray()
        while chunk := response.read(40960):
            content.extend(chunk)
        return content
    finally:
        conn.close()


def node_travel(node: Node, dfs: bool = True):
    """ usage example:
    for n in node_travel(node):
        print(n)
    """
    if not dfs:
        yield node
    for child in node.children:
        yield from node_travel(child, dfs)
    if dfs:
        yield node




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Mon Mar 04 2024 14:10:00 by codeskyblue
"""

from PIL import Image, ImageDraw

from uiautodev.driver.base_driver import BaseDriver
from uiautodev.model import Node, ShellResponse, WindowSize


class MockDriver(BaseDriver):
    def screenshot(self, id: int):
        im = Image.new("RGB", (500, 800), "gray")
        draw = ImageDraw.Draw(im)
        draw.text((10, 10), "mock", fill="white")
        draw.rectangle([100, 100, 200, 200], outline="red", fill="blue")
        del draw
        return im

    def dump_hierarchy(self):
        return "", Node(
            key="0",
            name="root",
            bounds=(0, 0, 1, 1),
            properties={
                "class": "android.view.View",
            },
            children=[
                Node(
                    key="0-0",
                    name="mock1",
                    bounds=(0.1, 0.1, 0.5, 0.5),
                    properties={
                        "class": "android.widget.FrameLayout",
                        "text": "mock1",
                        "accessible": "true",
                    },
                ),
                Node(
                    key="0-1",
                    name="mock2",
                    bounds=(0.4, 0.4, 0.6, 0.6),
                    properties={
                        "class": "android.widget.ImageView",
                        "text": "mock2",
                        "accessible": "true",
                    },
                    children=[
                        Node(
                            key="0-1-0",
                            name="mock2-1",
                            bounds=(0.42, 0.42, 0.45, 0.45),
                            properties={
                                "class": "android.widget.ImageView",
                                "text": "mock2-1",
                                "visible": "true",
                            },
                        ),
                    ]
                ),
                Node(
                    key="0-2",
                    name="mock-should-not-show",
                    bounds=(0.4, 0.4, 0.6, 0.6),
                    properties={
                        "class": "android.widget.ImageView",
                        "text": "mock3",
                        "visible": "false",
                    },
                ),
            ],
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Tue Mar 05 2024 11:16:29 by codeskyblue
"""

class UiautoException(Exception):
    pass


class IOSDriverException(UiautoException):
    pass


class AndroidDriverException(UiautoException):
    pass


class AppiumDriverException(UiautoException):
    pass


class MethodError(UiautoException):
    pass


class ElementNotFoundError(MethodError):
    pass


class RequestError(UiautoException):
    pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import json
import logging
import os
import re
import shutil
import subprocess
import tempfile
import time
import uuid
from pathlib import Path
from typing import List, Optional, Tuple, Union, final

from PIL import Image

from uiautodev.command_types import CurrentAppResponse
from uiautodev.driver.base_driver import BaseDriver
from uiautodev.model import AppInfo, Node, Rect, ShellResponse, WindowSize

logger = logging.getLogger(__name__)

StrOrPath = Union[str, Path]


def run_command(command: str, timeout: int = 60) -> str:
    try:
        result = subprocess.run(
            command,
            shell=True,
            capture_output=True,
            timeout=timeout,
            text=True,
            input='' # this avoid stdout: "FreeChannelContinue handle->data is nullptr"
        )
        # the hdc shell stderr is (不仅没啥用，还没办法去掉)
        # Remote PTY will not be allocated because stdin is not a terminal.
        # Use multiple -t options to force remote PTY allocation.
        output = result.stdout.strip()
        return output
    except subprocess.TimeoutExpired as e:
        raise TimeoutError(f"{command:r} timeout {e}")


class HDCError(Exception):
    pass


class HDC:
    def __init__(self):
        self.hdc = 'hdc'
        self.tmpdir = tempfile.TemporaryDirectory()
    
    def __del__(self):
        self.tmpdir.cleanup()
    
    def list_device(self) -> List[str]:
        command = f"{self.hdc} list targets"
        result = run_command(command)
        if result and not "Empty" in result:
            devices = []
            for line in result.strip().split("\n"):
                serial = line.strip().split('\t', 1)[0]
                devices.append(serial)
            return devices
        else:
            return []
    
    def shell(self, serial: str, command: str) -> str:
        command = f"{self.hdc} -t {serial} shell \"{command}\""
        result = run_command(command)
        return result.strip()
    
    def get_model(self, serial: str) -> str:
        return self.shell(serial, "param get const.product.model")
    
    def pull(self, serial: str, remote: StrOrPath, local: StrOrPath):
        if isinstance(remote, Path):
            remote = remote.as_posix()
        command = f"{self.hdc} -t {serial} file recv {remote} {local}"
        output = run_command(command)
        if not os.path.exists(local):
            raise HDCError(f"device file: {remote} not found", output)
    
    def push(self, serial: str, local: StrOrPath, remote: StrOrPath) -> str:
        if isinstance(remote, Path):
            remote = remote.as_posix()
        command = f"{self.hdc} -t {serial} file send {local} {remote}"
        return run_command(command)
    
    def screenshot(self, serial: str) -> Image.Image:
        device_path = f'/data/local/tmp/screenshot-{int(time.time()*1000)}.png'
        self.shell(serial, f"uitest screenCap -p {device_path}")
        try:
            local_path = os.path.join(self.tmpdir.name, f"{uuid.uuid4()}.png")
            self.pull(serial, device_path, local_path)
            with Image.open(local_path) as image:
                image.load()
                return image
        finally:
            self.shell(serial, f"rm {device_path}")

    def dump_layout(self, serial: str) -> dict:
        name = "{}.json".format(int(time.time() * 1000))
        remote_path = f"/data/local/tmp/layout-{name}.json"
        temp_path = os.path.join(self.tmpdir.name, f"layout-{name}.json")
        output = self.shell(serial, f"uitest dumpLayout -p {remote_path}")
        self.pull(serial, remote_path, temp_path)
        # mock
        # temp_path = Path(__file__).parent / 'testdata/layout.json'
        try:
            with open(temp_path, "rb") as f:
                json_content = json.load(f)
            return json_content
        except json.JSONDecodeError:
            raise HDCError(f"failed to dump layout: {output}")
        finally:
            self.shell(serial, f"rm {remote_path}")
    

class HarmonyDriver(BaseDriver):
    def __init__(self, hdc: HDC, serial: str):
        super().__init__(serial)
        self.hdc = hdc

    def screenshot(self, id: int = 0) -> Image.Image:
        return self.hdc.screenshot(self.serial)

    def window_size(self) -> WindowSize:
        result = self.hdc.shell(self.serial, "hidumper -s 10 -a screen")
        pattern = r"activeMode:\s*(\d+x\d+)"
        match = re.search(pattern, result)
        if match:
            resolution = match.group(1).split("x")
            return WindowSize(width=int(resolution[0]), height=int(resolution[1]))
        else:
            image = self.screenshot()
            return WindowSize(width=image.width, height=image.height)

    def dump_hierarchy(self) -> Tuple[str, Node]:
        """returns xml string and hierarchy object"""
        layout = self.hdc.dump_layout(self.serial)
        return json.dumps(layout), parse_json_element(layout)

    def tap(self, x: int, y: int):
        self.hdc.shell(self.serial, f"uinput -T -c {x} {y}")

    def app_current(self) -> Optional[CurrentAppResponse]:
        echo = self.hdc.shell(self.serial, "hidumper -s WindowManagerService -a '-a'")
        focus_window = re.search(r"Focus window: (\d+)", echo)
        if focus_window:
            focus_window = focus_window.group(1)
        mission_echo = self.hdc.shell(self.serial, "aa dump -a")
        pkg_names = re.findall(r"Mission ID #(\d+)\s+mission name #\[(.*?)\]", mission_echo)
        if focus_window and pkg_names:
            for mission in pkg_names:
                mission_id = mission[0]
                if focus_window == mission_id:
                    mission_name = mission[1]
                    pkg_name = mission_name.split(":")[0].replace("#", "")
                    ability_name = mission_name.split(":")[-1]
                    pid = self.hdc.shell(self.serial, f"pidof {pkg_name}").strip()
                    return CurrentAppResponse(package=pkg_name, activity=ability_name, pid=int(pid))
        else:
            return None

    def shell(self, command: str) -> ShellResponse:
        result = self.hdc.shell(self.serial, command)
        return ShellResponse(output=result)

    def home(self):
        self.hdc.shell(self.serial, "uinput -K -d 1 -u 1")

    def back(self):
        self.hdc.shell(self.serial, "uinput -K -d 2 -u 2")

    def volume_up(self):
        self.hdc.shell(self.serial, "uinput -K -d 16 -u 16")

    def volume_down(self):
        self.hdc.shell(self.serial, "uinput -K -d 17 -u 17")

    def volume_mute(self):
        self.hdc.shell(self.serial, "uinput -K -d 22 -u 22")

    def app_switch(self):
        self.hdc.shell(self.serial, "uinput -K -d 2076 -d 2049 -u 2076 -u 2049")

    def app_list(self) -> List[AppInfo]:
        results = []
        output = self.hdc.shell(self.serial, "bm dump -a")
        for i in output.split("\n"):
            if "ID" in i:
                continue
            else:
                results.append(AppInfo(packageName=i.strip()))
        return results


def parse_json_element(element, indexes: List[int] = [0]) -> Node:
    """
    Recursively parse an json element into a dictionary format.
    """
    attributes = element.get("attributes", {})
    name = attributes.get("type", "")
    bounds = attributes.get("bounds", "")
    bounds = list(map(int, re.findall(r"\d+", bounds)))
    assert len(bounds) == 4
    rect = Rect(x=bounds[0], y=bounds[1], width=bounds[2] - bounds[0], height=bounds[3] - bounds[1])
    elem = Node(
        key="-".join(map(str, indexes)),
        name=name,
        bounds=None,
        rect=rect,
        properties={key: attributes[key] for key in attributes},
        children=[],
    )
    # Construct xpath for children
    for index, child in enumerate(element.get("children", [])):
        child_node = parse_json_element(child, indexes + [index])
        if child_node:
            elem.children.append(child_node)

    return elem



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Tue Mar 05 2024 10:18:09 by codeskyblue

Copy from https://github.com/doronz88/pymobiledevice3
"""


from uiautodev.exceptions import IOSDriverException


class NotPairedError(IOSDriverException):
    pass



class MuxException(IOSDriverException):
    pass


class MuxVersionError(MuxException):
    pass


class BadCommandError(MuxException):
    pass


class BadDevError(MuxException):
    pass


class ConnectionFailedError(MuxException):
    pass


class ConnectionFailedToUsbmuxdError(ConnectionFailedError):
    pass


class ArgumentError(IOSDriverException):
    pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Tue Mar 19 2024 22:23:37 by codeskyblue
"""

import sys

import httpx
from fastapi import FastAPI, Request, Response

app = FastAPI()


# Retrieve the target URL from the command line arguments
try:
    TARGET_URL = sys.argv[1]
except IndexError:
    print("Usage: python proxy_server.py <target_url>")
    sys.exit(1)

@app.api_route("/{path:path}", methods=["GET", "POST", "PUT", "DELETE", "OPTIONS", "HEAD", "PATCH", "TRACE"])
async def proxy(request: Request, path: str):
    # Construct the full URL to forward the request to
    if path.endswith('/execute/sync'):
        # 旧版appium处理不好这个请求，直接返回404, unknown command
        # 目前browserstack也不支持这个请求
        return Response(content=b'{"value": {"error": "unknown command", "message": "unknown command", "stacktrace": "UnknownCommandError"}}', status_code=404)
    full_url = f"{TARGET_URL}/{path}"
    body = await request.body()
    print("Forwarding to", request.method, full_url)
    print("==> BODY <==")
    print(body)
    # Include original headers in the request
    headers = {k: v for k, v in request.headers.items() if k != 'host'}

    # Forward the request to the target server
    async with httpx.AsyncClient(timeout=120) as client:
        resp = await client.request(
            method=request.method,
            url=full_url,
            headers=headers,
            data=body,
            follow_redirects=True,        
        )

    # Return the response received from the target server
    return Response(content=resp.content, status_code=resp.status_code, headers=dict(resp.headers))


if __name__ == "__main__":
    import uvicorn
    uvicorn.run(app, host="0.0.0.0", port=8000)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Fri Mar 01 2024 14:35:46 by codeskyblue
"""


import json
from functools import partial
from typing import List, Optional, Tuple
from xml.etree import ElementTree

import wdapy
from PIL import Image

from uiautodev.command_types import CurrentAppResponse
from uiautodev.driver.base_driver import BaseDriver
from uiautodev.exceptions import IOSDriverException
from uiautodev.model import Node, WindowSize
from uiautodev.utils.usbmux import select_device


class IOSDriver(BaseDriver):
    def __init__(self, serial: str):
        """ serial is the udid of the ios device """
        super().__init__(serial)
        self.device = select_device(serial)
        self.wda = wdapy.AppiumUSBClient(self.device.serial)
    
    def _request(self, method: str, path: str, payload: Optional[dict] = None) -> bytes:
        conn = self.device.make_http_connection(port=8100)
        try:
            if payload is None:
                conn.request(method, path)
            else:
                conn.request(method, path, body=json.dumps(payload), headers={"Content-Type": "application/json"})
            response = conn.getresponse()
            if response.getcode() != 200:
                raise IOSDriverException(f"Failed request to device, status: {response.getcode()}")
            content = bytearray()
            while chunk := response.read(4096):
                content.extend(chunk)
            return content
        finally:
            conn.close()
    
    def _request_json(self, method: str, path: str) -> dict:
        content = self._request(method, path)
        return json.loads(content)

    def _request_json_value(self, method: str, path: str) -> dict:
        return self._request_json(method, path)["value"]
    
    def status(self):
        return self._request_json("GET", "/status")
    
    def screenshot(self, id: int = 0) -> Image.Image:
        return self.wda.screenshot()
    
    def window_size(self):
        return self.wda.window_size()
    
    def dump_hierarchy(self) -> Tuple[str, Node]:
        """returns xml string and hierarchy object"""
        t = self.wda.sourcetree()
        xml_data = t.value
        root = ElementTree.fromstring(xml_data)
        return xml_data, parse_xml_element(root, WindowSize(width=1, height=1))
    
    def tap(self, x: int, y: int):
        self.wda.tap(x, y)
    
    def app_current(self) -> CurrentAppResponse:
        info = self.wda.app_current()
        return CurrentAppResponse(package=info.bundle_id, pid=info.pid)

    def home(self):
        self.wda.homescreen()
    
    def app_switch(self):
        raise NotImplementedError()

    def volume_up(self):
        self.wda.volume_up()
    
    def volume_down(self):
        self.wda.volume_down()
        

def parse_xml_element(element, wsize: WindowSize, indexes: List[int]=[0]) -> Node:
    """
    Recursively parse an XML element into a dictionary format.
    # <XCUIElementTypeApplication type="XCUIElementTypeApplication" name="设置" label="设置" enabled="true" visible="true" accessible="false" x="0" y="0" width="414" height="896" index="0">
    """
    if element.attrib.get("visible") == "false":
        return None
    if element.tag == "XCUIElementTypeApplication":
        wsize = WindowSize(width=int(element.attrib["width"]), height=int(element.attrib["height"]))
    x = int(element.attrib.get("x", 0))
    y = int(element.attrib.get("y", 0))
    width = int(element.attrib.get("width", 0))
    height = int(element.attrib.get("height", 0))
    bounds = (x / wsize.width, y / wsize.height, (x + width) / wsize.width, (y + height) / wsize.height)
    bounds = list(map(partial(round, ndigits=4), bounds))
    name = element.attrib.get("type", "XCUIElementTypeUnknown")
    
    elem = Node(
        key='-'.join(map(str, indexes)),
        name=name,
        bounds=bounds,
        properties={key: element.attrib[key] for key in element.attrib},
        children=[],
    )
    for index, child in enumerate(element):
        child_elem = parse_xml_element(child, wsize, indexes+[index])
        if child_elem:
            elem.children.append(child_elem)
    return elem

    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Fri Mar 01 2024 14:19:29 by codeskyblue
"""

import logging
import re
import time
from functools import cached_property, partial
from typing import Iterator, List, Optional, Tuple
from xml.etree import ElementTree

import adbutils
import uiautomator2 as u2
from PIL import Image

from uiautodev.command_types import CurrentAppResponse
from uiautodev.driver.base_driver import BaseDriver
from uiautodev.exceptions import AndroidDriverException, RequestError
from uiautodev.model import AppInfo, Node, Rect, ShellResponse, WindowSize
from uiautodev.utils.common import fetch_through_socket

logger = logging.getLogger(__name__)

class AndroidDriver(BaseDriver):
    def __init__(self, serial: str):
        super().__init__(serial)
        self.adb_device = adbutils.device(serial)

    @cached_property
    def ud(self) -> u2.Device:
        return u2.connect_usb(self.serial)
    
    def screenshot(self, id: int) -> Image.Image:
        if id > 0:
            raise AndroidDriverException("multi-display is not supported yet for uiautomator2")
        return self.ud.screenshot()

    def shell(self, command: str) -> ShellResponse:
        try:
            ret = self.adb_device.shell2(command, rstrip=True, timeout=20)
            if ret.returncode == 0:
                return ShellResponse(output=ret.output, error=None)
            else:
                return ShellResponse(
                    output="", error=f"exit:{ret.returncode}, output:{ret.output}"
                )
        except Exception as e:
            return ShellResponse(output="", error=f"adb error: {str(e)}")

    def dump_hierarchy(self, display_id: Optional[int] = 0) -> Tuple[str, Node]:
        """returns xml string and hierarchy object"""
        start = time.time()
        xml_data = self._dump_hierarchy_raw()
        logger.debug("dump_hierarchy cost: %s", time.time() - start)

        wsize = self.adb_device.window_size()
        logger.debug("window size: %s", wsize)
        return xml_data, parse_xml(
            xml_data, WindowSize(width=wsize[0], height=wsize[1]), display_id
        )

    def _dump_hierarchy_raw(self) -> str:
        """
        uiautomator2 server is conflict with "uiautomator dump" command.

        uiautomator dump errors:
        - ERROR: could not get idle state.
        """
        try:
            return self.ud.dump_hierarchy()
        except Exception as e:
            raise AndroidDriverException(f"Failed to dump hierarchy: {str(e)}")
    
    def tap(self, x: int, y: int):
        self.adb_device.click(x, y)

    def window_size(self) -> Tuple[int, int]:
        w, h = self.adb_device.window_size()
        return (w, h)

    def app_install(self, app_path: str):
        self.adb_device.install(app_path)

    def app_current(self) -> CurrentAppResponse:
        info = self.adb_device.app_current()
        return CurrentAppResponse(
            package=info.package, activity=info.activity, pid=info.pid
        )

    def app_launch(self, package: str):
        if self.adb_device.package_info(package) is None:
            raise AndroidDriverException(f"App not installed: {package}")
        self.adb_device.app_start(package)
    
    def app_terminate(self, package: str):
        self.adb_device.app_stop(package)

    def home(self):
        self.adb_device.keyevent("HOME")
    
    def wake_up(self):
        self.adb_device.keyevent("WAKEUP")
    
    def back(self):
        self.adb_device.keyevent("BACK")
    
    def app_switch(self):
        self.adb_device.keyevent("APP_SWITCH")
    
    def volume_up(self):
        self.adb_device.keyevent("VOLUME_UP")
    
    def volume_down(self):
        self.adb_device.keyevent("VOLUME_DOWN")
    
    def volume_mute(self):
        self.adb_device.keyevent("VOLUME_MUTE")
    
    def app_list(self) -> List[AppInfo]:
        results = []
        output = self.adb_device.shell(["pm", "list", "packages", '-3'])
        for m in re.finditer(r"^package:([^\s]+)\r?$", output, re.M):
            packageName = m.group(1)
            results.append(AppInfo(packageName=packageName))
        return results

    def open_app_file(self, package: str) -> Iterator[bytes]:
        line = self.adb_device.shell(f"pm path {package}")
        if not line.startswith("package:"):
            raise AndroidDriverException(f"Failed to get package path: {line}")
        remote_path = line.split(':', 1)[1]
        yield from self.adb_device.sync.iter_content(remote_path)



def parse_xml(xml_data: str, wsize: WindowSize, display_id: Optional[int] = None) -> Node:
    root = ElementTree.fromstring(xml_data)
    node = parse_xml_element(root, wsize, display_id)
    if node is None:
        raise AndroidDriverException("Failed to parse xml")
    return node


def parse_xml_element(element, wsize: WindowSize, display_id: Optional[int], indexes: List[int] = [0]) -> Optional[Node]:
    """
    Recursively parse an XML element into a dictionary format.
    """
    name = element.tag
    if name == "node":
        name = element.attrib.get("class", "node")
    if display_id is not None:
        elem_display_id = int(element.attrib.get("display-id", display_id))
        if elem_display_id != display_id:
            return

    bounds = None
    rect = None
    # eg: bounds="[883,2222][1008,2265]"
    if "bounds" in element.attrib:
        bounds = element.attrib["bounds"]
        bounds = list(map(int, re.findall(r"\d+", bounds)))
        assert len(bounds) == 4
        rect = Rect(x=bounds[0], y=bounds[1], width=bounds[2] - bounds[0], height=bounds[3] - bounds[1])
        bounds = (
            bounds[0] / wsize.width,
            bounds[1] / wsize.height,
            bounds[2] / wsize.width,
            bounds[3] / wsize.height,
        )
        bounds = map(partial(round, ndigits=4), bounds)
        
    elem = Node(
        key="-".join(map(str, indexes)),
        name=name,
        bounds=bounds,
        rect=rect,
        properties={key: element.attrib[key] for key in element.attrib},
        children=[],
    )

    # Construct xpath for children
    for index, child in enumerate(element):
        child_node = parse_xml_element(child, wsize, display_id, indexes + [index])
        if child_node:
            elem.children.append(child_node)

    return elem



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Sun Apr 21 2024 21:15:15 by codeskyblue
"""


import atexit
import enum
import io
import json
import logging
import threading
import time
from base64 import b64decode
from pathlib import Path
from pprint import pprint
from typing import Any, Optional

import adbutils
import requests
from PIL import Image
from pydantic import BaseModel

"""
shell steps:
adb push appium-uiautomator2-v5.12.4.apk /data/local/tmp/udt.jar
adb shell CLASSPATH=/data/local/tmp/udt.jar app_process / "com.wetest.uia2.Main"
adb forward tcp:6790 tcp:6790
# 创建session
echo '{"capabilities": {}}' | http POST :6790/session
# 获取当前所有session
http GET :6790/sessions
# 获取pageSource
http GET :6790/session/{session_id}/source

# TODO
# /appium/settins 中waitForIdleTimeout需要调整，其他的再看看
"""

logger = logging.getLogger(__name__)

class UDTError(Exception):
    pass


class HTTPError(UDTError):
    pass


class AppiumErrorEnum(str, enum.Enum):
    InvalidSessionID = 'invalid session id'


class AppiumError(UDTError):
    def __init__(self, error: str, message):
        self.error = error
        self.message = message


class AppiumResponseValue(BaseModel):
    error: Optional[str] = None
    message: Optional[str] = None
    stacktrace: Optional[str] = None


class AppiumResponse(BaseModel):
    sessionId: Optional[str] = None
    value: Any = None


class MockAdbProcess:
    def __init__(self, conn: adbutils.AdbConnection) -> None:
        self._conn = conn
        self._event = threading.Event()
        
        def wait_finished():
            try:
                self._conn.read_until_close()
            except:
                pass
            self._event.set()
        
        t = threading.Thread(target=wait_finished)
        t.daemon = True
        t.name = "wait_adb_conn"
        t.start()

    def wait(self) -> int:
        self._event.wait()
        return 0

    def pool(self) -> Optional[int]:
        if self._event.is_set():
            return 0
        return None

    def kill(self):
        self._conn.close()


class UDT:
    def __init__(self, device: adbutils.AdbDevice):
        self._device = device
        self._lport = None
        self._process = None
        self._lock = threading.Lock()
        self._session_id = None
        atexit.register(self.release)

    def get_session_id(self) -> str:
        if self._session_id:
            return self._session_id
        self._session_id = self._new_session()
        logger.debug("update waitForIdleTimeout to 0ms")
        self._dev_request("POST", f"/session/{self._session_id}/appium/settings", payload={
            "settings": {
                "waitForIdleTimeout": 10,
                "waitForSelectorTimeout": 10,
                "actionAcknowledgmentTimeout": 10,
                "scrollAcknowledgmentTimeout": 10,
                "trackScrollEvents": False,
            }
        })
        result = self._dev_request("GET", f"/session/{self._session_id}/appium/settings")
        return self._session_id
    
    def dev_request(self, method: str, path: str, **kwargs) -> AppiumResponse:
        """send http request to device
        :param method: GET, POST, DELETE, PUT
        :param path: url path, path start with @ means with_session=True

        :return: response json
        """
        try:
            if path.startswith("@"):
                path = path[1:]
                kwargs['with_session'] = True
            return self._dev_request(method, path, **kwargs)
        except HTTPError:
            self.launch_server()
            return self._dev_request(method, path, **kwargs)
        except AppiumError as e:
            if e.error == AppiumErrorEnum.InvalidSessionID:
                self._session_id = self._new_session()
                return self._dev_request(method, path, **kwargs)
            raise

    def _dev_request(self, method: str, path: str, payload=None, timeout: float = 10.0, with_session: bool = False) -> AppiumResponse:
        try:
            if with_session:
                sid = self.get_session_id()
                path = f"/session/{sid}{path}"
            url = f"http://localhost:{self._lport}{path}"
            logger.debug("request %s %s", method, url)
            r = requests.request(method, url, json=payload, timeout=timeout)
            response_json = r.json()
            resp = AppiumResponse.model_validate(response_json)
            if isinstance(resp.value, dict):
                value = AppiumResponseValue.model_validate(resp.value)
                if value.error:
                    raise AppiumError(value.error, value.message)
            return resp
        except requests.RequestException as e:
            raise HTTPError(f"{method} to {path!r} error", payload)
        except json.JSONDecodeError as e:
            raise HTTPError("JSON decode error", e.msg)
        
    def _new_session(self) -> str:
        resp = self._dev_request("POST", "/session", payload={"capabilities": {}})
        value = resp.value
        if not isinstance(value, dict) and 'sessionId' not in value:
            raise UDTError("session create failed", resp)
        sid = value['sessionId']
        if not sid:
            raise UDTError("session create failed", resp)
        return sid

    def post(self, path: str, payload=None) -> AppiumResponse:
        return self.dev_request("POST", path, payload=payload)

    def get(self, path: str, ) -> AppiumResponse:
        return self.dev_request("GET", path)

    def _update_process_status(self):
        if self._process:
            if self._process.pool() is not None:
                self._process = None

    def release(self):
        logger.debug("Releasing")
        with self._lock:
            if self._process is not None:
                logger.debug("Killing process")
                self._process.kill()
                self._process.wait()
                self._process = None

    def launch_server(self):
        try:
            self._launch_server()
            self._device.keyevent("WAKEUP")
        except adbutils.AdbError as e:
            raise UDTError("fail to start udt", str(e))
        self._wait_ready()

    def _launch_server(self):
        with self._lock:
            self._update_process_status()
            if self._process:
                logger.debug("Process already running")
                return
            logger.debug("Launching process")
            dex_local_path = Path(__file__).parent.joinpath("appium-uiautomator2-v5.12.4-light.apk")
            logger.debug("dex_local_path: %s", dex_local_path)
            dex_remote_path = "/data/local/tmp/udt/udt-5.12.4-light.dex"
            info = self._device.sync.stat(dex_remote_path)
            if info.size == dex_local_path.stat().st_size:
                logger.debug("%s already exists", dex_remote_path)
            else:
                logger.debug("push dex(%d) to %s", dex_local_path.stat().st_size, dex_remote_path)
                self._device.shell("mkdir -p /data/local/tmp/udt")
                self._device.sync.push(dex_local_path, dex_remote_path, 0o644)
            logger.debug("CLASSPATH=%s app_process / com.wetest.uia2.Main", dex_remote_path)
            conn = self._device.shell(f"CLASSPATH={dex_remote_path} app_process / com.wetest.uia2.Main", stream=True)
            self._process = MockAdbProcess(conn)

            self._lport = self._device.forward_port(6790)
            logger.debug("forward tcp:6790 -> tcp:%d", self._lport)

    def _wait_ready(self):
        deadline = time.time() + 10
        while time.time() < deadline:
            try:
                self._dev_request("GET", "/status", timeout=1)
                return
            except HTTPError:
                time.sleep(0.5)
        raise UDTError("Service not ready")

    def dump_hierarchy(self) -> str:
        resp = self.get(f"@/source")
        return resp.value
    
    def status(self):
        return self.get("/status")
    
    def screenshot(self) -> Image.Image:
        resp = self.get(f"@/screenshot")
        raw = b64decode(resp.value)
        return Image.open(io.BytesIO(raw))



if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    r = UDT(adbutils.device())
    print(r.status())
    r.dump_hierarchy()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Sun Feb 18 2024 11:10:58 by codeskyblue
"""
from __future__ import annotations

import abc
from functools import lru_cache

import adbutils

from uiautodev.driver.android import AndroidDriver
from uiautodev.driver.base_driver import BaseDriver
from uiautodev.driver.harmony import HDC, HarmonyDriver
from uiautodev.driver.ios import IOSDriver
from uiautodev.driver.mock import MockDriver
from uiautodev.exceptions import UiautoException
from uiautodev.model import DeviceInfo
from uiautodev.utils.usbmux import MuxDevice, list_devices


class BaseProvider(abc.ABC):
    @abc.abstractmethod
    def list_devices(self) -> list[DeviceInfo]:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_device_driver(self, serial: str) -> BaseDriver:
        raise NotImplementedError()

    def get_single_device_driver(self) -> BaseDriver:
        """ debug use """
        devs = self.list_devices()
        if len(devs) == 0:
            raise UiautoException("No device found")
        if len(devs) > 1:
            raise UiautoException("More than one device found")
        return self.get_device_driver(devs[0].serial)


class AndroidProvider(BaseProvider):
    def __init__(self):
        pass

    def list_devices(self) -> list[DeviceInfo]:
        adb = adbutils.AdbClient()
        ret: list[DeviceInfo] = []
        for d in adb.list():
            if d.state != "device":
                ret.append(DeviceInfo(serial=d.serial, status=d.state, enabled=False))
            else:
                dev = adb.device(d.serial)
                ret.append(DeviceInfo(serial=d.serial, model=dev.prop.model, name=dev.prop.name))
        return ret

    @lru_cache
    def get_device_driver(self, serial: str) -> AndroidDriver:
        return AndroidDriver(serial)


class IOSProvider(BaseProvider):
    def list_devices(self) -> list[DeviceInfo]:
        devs = list_devices()
        return [DeviceInfo(serial=d.serial, model="unknown", name="unknown") for d in devs]

    @lru_cache
    def get_device_driver(self, serial: str) -> BaseDriver:
        return IOSDriver(serial)


class HarmonyProvider(BaseProvider):
    def __init__(self):
        super().__init__()
        self.hdc = HDC()

    def list_devices(self) -> list[DeviceInfo]:
        devices = self.hdc.list_device()
        return [DeviceInfo(serial=d, model=self.hdc.get_model(d), name=self.hdc.get_model(d)) for d in devices]

    @lru_cache
    def get_device_driver(self, serial: str) -> HarmonyDriver:
        return HarmonyDriver(self.hdc, serial)


class MockProvider(BaseProvider):
    def list_devices(self) -> list[DeviceInfo]:
        return [DeviceInfo(serial="mock-serial", model="mock-model", name="mock-name")]

    def get_device_driver(self, serial: str) -> BaseDriver:
        return MockDriver(serial)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Tue Mar 19 2024 10:19:27 by codeskyblue
"""


# Request and Response
import enum
from typing import List, Optional, Union

from pydantic import BaseModel

from uiautodev.model import Node


# POST /api/v1/device/{serial}/command/{command}
class Command(str, enum.Enum):
    TAP = "tap"
    TAP_ELEMENT = "tapElement"
    APP_INSTALL = "installApp"
    APP_CURRENT = "currentApp"
    APP_LAUNCH = "appLaunch"
    APP_TERMINATE = "appTerminate"
    APP_LIST = "appList"

    GET_WINDOW_SIZE = "getWindowSize"
    HOME = "home"
    DUMP = "dump"
    WAKE_UP = "wakeUp"
    FIND_ELEMENTS = "findElements"
    CLICK_ELEMENT = "clickElement"

    LIST = "list"

    # 0.4.0
    BACK = "back"
    APP_SWITCH = "appSwitch"
    VOLUME_UP = "volumeUp"
    VOLUME_DOWN = "volumeDown"
    VOLUME_MUTE = "volumeMute"


class TapRequest(BaseModel):
    x: Union[int, float]
    y: Union[int, float]
    isPercent: bool = False


class InstallAppRequest(BaseModel):
    url: str


class InstallAppResponse(BaseModel):
    success: bool
    id: Optional[str] = None


class CurrentAppResponse(BaseModel):
    package: str
    activity: Optional[str] = None
    pid: Optional[int] = None


class AppLaunchRequest(BaseModel):
    package: str
    stop: bool = False


class AppTerminateRequest(BaseModel):
    package: str


class WindowSizeResponse(BaseModel):
    width: int
    height: int


class DumpResponse(BaseModel):
    value: str


class By(str, enum.Enum):
    ID = "id"
    TEXT = "text"
    XPATH = "xpath"
    CLASS_NAME = "className"

class FindElementRequest(BaseModel):
    by: str
    value: str
    timeout: float = 10.0


class FindElementResponse(BaseModel):
    count: int
    value: List[Node]


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Fri Mar 01 2024 14:18:30 by codeskyblue
"""
import abc
from io import FileIO
from typing import Iterator, List, Tuple

from PIL import Image
from pydantic import BaseModel

from uiautodev.command_types import CurrentAppResponse
from uiautodev.model import AppInfo, Node, ShellResponse, WindowSize


class BaseDriver(abc.ABC):
    def __init__(self, serial: str):
        self.serial = serial

    @abc.abstractmethod
    def screenshot(self, id: int) -> Image.Image:
        """Take a screenshot of the device
        :param id: physical display ID to capture (normally: 0)
        :return: PIL.Image.Image
        """
        raise NotImplementedError()
    
    @abc.abstractmethod
    def dump_hierarchy(self) -> Tuple[str, Node]:
        """Dump the view hierarchy of the device
        :return: xml_source, Hierarchy
        """
        raise NotImplementedError()
    
    def shell(self, command: str) -> ShellResponse:
        """Run a shell command on the device
        :param command: shell command
        :return: ShellResponse
        """
        raise NotImplementedError()
    
    def tap(self, x: int, y: int):
        """Tap on the screen
        :param x: x coordinate
        :param y: y coordinate
        """
        raise NotImplementedError()

    def window_size(self) -> WindowSize:
        """ get window UI size """
        raise NotImplementedError()

    def app_install(self, app_path: str):
        """ install app """
        raise NotImplementedError()
    
    def app_current(self) -> CurrentAppResponse:
        """ get current app """
        raise NotImplementedError()
    
    def app_launch(self, package: str):
        """ launch app """
        raise NotImplementedError()
    
    def app_terminate(self, package: str):
        """ terminate app """
        raise NotImplementedError()
    
    def home(self):
        """ press home button """
        raise NotImplementedError()
    
    def back(self):
        """ press back button """
        raise NotImplementedError()

    def app_switch(self):
        """ switch app """
        raise NotImplementedError()
    
    def volume_up(self):
        """ volume up """
        raise NotImplementedError()
    
    def volume_down(self):
        """ volume down """
        raise NotImplementedError()
    
    def volume_mute(self):
        """ volume mute """
        raise NotImplementedError()

    def wake_up(self):
        """ wake up the device """
        raise NotImplementedError()
    
    def app_list(self) -> List[AppInfo]:
        """ list installed packages """
        raise NotImplementedError()
    
    def open_app_file(self, package: str) -> Iterator[bytes]:
        """ open app file """
        raise NotImplementedError()
        


import enum
import socket
import struct

from construct import Byte, Int16ub, Int32ub, Int64ub, Struct

from uiautodev.remote.android_input import KeyeventAction, MetaState
from uiautodev.remote.keycode import KeyCode


# https://github.com/Genymobile/scrcpy/blob/master/app/src/control_msg.h#L29
class MessageType(enum.IntEnum):
    INJECT_KEYCODE = 0
    INJECT_TEXT = 1
    INJECT_TOUCH_EVENT = 2
    INJECT_SCROLL_EVENT = 3
    BACK_OR_SCREEN_ON = 4
    EXPAND_NOTIFICATION_PANEL = 5
    EXPAND_SETTINGS_PANEL = 6
    COLLAPSE_PANELS = 7
    GET_CLIPBOARD = 8
    SET_CLIPBOARD = 9
    SET_DISPLAY_POWER = 10
    ROTATE_DEVICE = 11
    UHID_CREATE = 12
    UHID_INPUT = 13
    UHID_DESTROY = 14
    OPEN_HARD_KEYBOARD_SETTINGS = 15
    START_APP = 16
    RESET_VIDEO = 17


TouchEvent = Struct(
    "type" / Byte,               # SC_CONTROL_MSG_TYPE_INJECT_TOUCH_EVENT
    "action" / Byte,             # AKEY_EVENT_ACTION_DOWN
    "pointer_id" / Int64ub,      # 8-byte pointer ID
    "x" / Int32ub,               # X coordinate
    "y" / Int32ub,               # Y coordinate
    "width" / Int16ub,           # width
    "height" / Int16ub,          # height
    "pressure" / Int16ub,        # pressure
    "action_button" / Int32ub,   # action button
    "buttons" / Int32ub          # buttons
)


# Define the structure for key events
KeyEvent = Struct(
    "type" / Byte,               # SC_CONTROL_MSG_TYPE_INJECT_KEYCODE
    "action" / Byte,             # AKEY_EVENT_ACTION (DOWN, UP, MULTIPLE)
    "keycode" / Int32ub,         # Android keycode
    "repeat" / Int32ub,          # Repeat count
    "metastate" / Int32ub        # Meta state flags (SHIFT, ALT, etc.)
)


class ScrcpyTouchController:
    """scrcpy控制类，支持scrcpy版本>=2.2"""

    def __init__(self, control_socket: socket.socket):
        self.control_socket = control_socket

    def _build_touch_event(self, action: int, x: int, y: int, width: int, height: int):
        x = max(0, min(x, width))
        y = max(0, min(y, height))
        return TouchEvent.build(dict(
            type=MessageType.INJECT_TOUCH_EVENT,
            action=action,
            pointer_id=1,
            x=x,
            y=y,
            width=width,
            height=height,
            pressure=1,
            action_button=1, # AMOTION_EVENT_BUTTON_PRIMARY (action button)
            buttons=1, # AMOTION_EVENT_BUTTON_PRIMARY (buttons)
        ))

    def down(self, x: int, y: int, width: int, height: int):
        """发送down操作"""
        data = self._build_touch_event(0, x, y, width, height)
        self.control_socket.send(data)

    def up(self, x: int, y: int, width: int, height: int):
        """发送up操作"""
        data = self._build_touch_event(1, x, y, width, height)
        self.control_socket.send(data)

    def move(self, x: int, y: int, width: int, height: int):
        """发送move操作"""
        data = self._build_touch_event(2, x, y, width, height)
        self.control_socket.send(data)

    def text(self, text: str):
        """发送文本操作"""

        # buffer = text.encode("utf-8")
        # values = struct.pack(self.format_string, 2, 3, 1, len(buffer), 0, 0, 0, self.const_value,
        #                      self.unknown1, self.unknown2) + buffer
        # self.control_socket.send(values)
        pass

    def key(self, action: KeyeventAction, keycode: KeyCode, repeat: int, metastate: MetaState):
        """
        Send a keycode event to the Android device
        
        Args:
            action: Key action (DOWN, UP, or MULTIPLE)
            keycode: Android key code to send
            repeat: Number of times the key is repeated
            metastate: Meta state flags (SHIFT, ALT, etc.)
        """
        # Build the data using the KeyEvent structure
        data = KeyEvent.build(dict(
            type=MessageType.INJECT_KEYCODE,  # Type byte
            action=action,                   # Action byte
            keycode=keycode,                   # Keycode (4 bytes)
            repeat=repeat,                   # Repeat count (4 bytes)
            metastate=metastate,        # Meta state (4 bytes)
        ))
        
        # Send the data to the control socket
        self.control_socket.send(data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Sun Feb 18 2024 11:12:33 by codeskyblue
"""
from __future__ import annotations

import typing
from typing import Dict, List, Optional, Tuple, Union

from pydantic import BaseModel


class DeviceInfo(BaseModel):
    serial: str
    model: str = ""
    name: str = ""
    status: str = ""
    enabled: bool = True


class ShellResponse(BaseModel):
    output: str
    error: Optional[str] = ""


class Rect(BaseModel):
    x: int
    y: int
    width: int
    height: int


class Node(BaseModel):
    key: str
    name: str # can be seen as description
    bounds: Optional[Tuple[float, float, float, float]] = None
    rect: Optional[Rect] = None
    properties: Dict[str, Union[str, bool]] = {}
    children: List[Node] = []


class OCRNode(Node):
    confidence: float


class WindowSize(typing.NamedTuple):
    width: int
    height: int


class AppInfo(BaseModel):
    packageName: str


# Ref
# https://github.com/Genymobile/scrcpy/blob/master/app/src/android/input.h
from enum import IntEnum


class MetaState(IntEnum):
    """Android meta state flags ported from Android's KeyEvent class
    
    These flags represent the state of meta keys such as ALT, SHIFT, CTRL, etc.
    They can be combined using bitwise OR operations to represent multiple
    meta keys being pressed simultaneously.
    
    The values and comments are taken directly from the Android source code
    to maintain compatibility and provide accurate descriptions.
    """
    # No meta keys are pressed
    NONE = 0x0
    
    # This mask is used to check whether one of the SHIFT meta keys is pressed
    SHIFT_ON = 0x1
    
    # This mask is used to check whether one of the ALT meta keys is pressed
    ALT_ON = 0x2
    
    # This mask is used to check whether the SYM meta key is pressed
    SYM_ON = 0x4
    
    # This mask is used to check whether the FUNCTION meta key is pressed
    FUNCTION_ON = 0x8
    
    # This mask is used to check whether the left ALT meta key is pressed
    ALT_LEFT_ON = 0x10
    
    # This mask is used to check whether the right ALT meta key is pressed
    ALT_RIGHT_ON = 0x20
    
    # This mask is used to check whether the left SHIFT meta key is pressed
    SHIFT_LEFT_ON = 0x40
    
    # This mask is used to check whether the right SHIFT meta key is pressed
    SHIFT_RIGHT_ON = 0x80
    
    # This mask is used to check whether the CAPS LOCK meta key is on
    CAPS_LOCK_ON = 0x100000
    
    # This mask is used to check whether the NUM LOCK meta key is on
    NUM_LOCK_ON = 0x200000
    
    # This mask is used to check whether the SCROLL LOCK meta key is on
    SCROLL_LOCK_ON = 0x400000
    
    # This mask is used to check whether one of the CTRL meta keys is pressed
    CTRL_ON = 0x1000
    
    # This mask is used to check whether the left CTRL meta key is pressed
    CTRL_LEFT_ON = 0x2000
    
    # This mask is used to check whether the right CTRL meta key is pressed
    CTRL_RIGHT_ON = 0x4000
    
    # This mask is used to check whether one of the META meta keys is pressed
    META_ON = 0x10000
    
    # This mask is used to check whether the left META meta key is pressed
    META_LEFT_ON = 0x20000
    
    # This mask is used to check whether the right META meta key is pressed
    META_RIGHT_ON = 0x40000


class KeyeventAction(IntEnum):
    DOWN = 0
    UP = 1
    MULTIPLE = 2



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Tue Mar 05 2024 16:59:19 by codeskyblue
"""

from fastapi import APIRouter, Form, Response
from lxml import etree
from typing_extensions import Annotated

router = APIRouter()


@router.post("/check/xpath")
def check_xpath(xml: Annotated[str, Form()], xpath: Annotated[str, Form()]) -> Response:
    """Check if the XPath expression is valid"""
    try:
        children = []
        for child in etree.fromstring(xml).xpath(xpath):
            children.append(child)
        if len(children) > 0:
            return Response(content=children[0].tag, media_type="text/plain")
        else:
            return Response(
                content="XPath is valid but not node matches", media_type="text/plain"
            )
    except Exception as e:
        return Response(content=str(e), media_type="text/plain", status_code=400)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Sat Apr 13 2024 22:35:03 by codeskyblue
"""

import enum
import logging
from typing import Dict, Union

from pydantic import BaseModel

from uiautodev import command_proxy
from uiautodev.command_types import Command
from uiautodev.driver.base_driver import BaseDriver
from uiautodev.provider import AndroidProvider

logger = logging.getLogger(__name__)

class CommandStep(BaseModel):
    method: Union[str, Command]
    params: Dict[str, str]
    skip: bool = False
    ignore_error: bool = False


class CompareEnum(str, enum.Enum):
    EQUAL = "equal"
    CONTAINS = "contains"
    NOT_EQUAL = "not_equal"
    NOT_CONTAINS = "not_contains"


class CompareCheckStep(BaseModel):
    method: CompareEnum
    value_a: str
    value_b: str
    skip: bool = False


def run_driver_command(driver: BaseDriver, command: Command, params: dict):
    model = command_proxy.get_command_params_type(command)
    params_obj = model.model_validate(params) if params else None
    # print("Params:", params, params_obj)
    result = command_proxy.send_command(driver, command, params_obj)
    return result

def run():
    # all params key and value should be string
    # Step中
    # 入参类型在为前端保存一份，后端需要同步兼容
    # params所有的key和value都是string类型
    # 出参类型支持重命名 result
    # - key: string, old_key: string, desc: string
    # eg. "WIDTH", "width", "屏幕宽度"
    steps = [
        CommandStep(
            method=Command.APP_LAUNCH,
            params= {
                "package": "com.saucelabs.mydemoapp.android",
                "stop": "true" # bool
            }
        ),
        CommandStep(
            method=Command.GET_WINDOW_SIZE,
            result_trans=[
                dict(key="WIDTH", result_key="width", desc="屏幕宽度"),
            ]
        ),
        CommandStep(
            method=Command.ECHO,
            params={
                "message": "WindowWidth is {{WIDTH}}", 
            }
        ),
        CommandStep(
            method=Command.CLICK_ELEMENT,
            params={
                "by": "id",
                "value": "com.saucelabs.mydemoapp.android:id/productIV",
            }
        ),
        CommandStep(
            method=Command.CLICK_ELEMENT,
            params={
                "by": "id",
                "value": "com.saucelabs.mydemoapp.android:id/plusIV",
            }
        ),
        CommandStep(
            method=Command.CLICK_ELEMENT,
            params={
                "by": "id",
                "value": "com.saucelabs.mydemoapp.android:id/cartBt",
            }
        ),
        CommandStep(
            method=Command.CLICK_ELEMENT,
            params={
                "by": "id",
                "value": "com.saucelabs.mydemoapp.android:id/cartIV",
            }
        ),
        CommandStep(
            method=Command.FIND_ELEMENT,
            params={
                "by": "text",
                "value": "Proceed To Checkout",
            },
            skip=True,
        ),
        CompareCheckStep(
            method=CompareEnum.EQUAL,
            value_a="$.name",
            value_b="com.saucelabs.mydemoapp.android:id/cartIV",
        ),
        CommandStep(
            method=Command.CLICK_ELEMENT,
            params={
                "by": "text",
                "value": "Proceed To Checkout",
            }
        ),
    ]
    provider = AndroidProvider()
    driver = provider.get_single_device_driver()
    local_vars: Dict[str, str] = {}
    for step in steps:
        if not isinstance(step, CommandStep):
            continue
        command = Command(step.method)
        params = step.params
        print(step.method, params)
        if step.skip:
            logger.debug("Skip step: %s", step.method)
            continue
        run_driver_command(driver, command, params)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Sun Feb 18 2024 13:48:55 by codeskyblue
"""

import logging
import os
import platform
import signal
from pathlib import Path
from typing import List

import adbutils
import uvicorn
from fastapi import FastAPI, File, UploadFile, WebSocket
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import FileResponse, JSONResponse, RedirectResponse
from pydantic import BaseModel
from rich.logging import RichHandler
from starlette.websockets import WebSocketDisconnect

from uiautodev import __version__
from uiautodev.common import convert_bytes_to_image, get_webpage_url, ocr_image
from uiautodev.model import Node
from uiautodev.provider import AndroidProvider, HarmonyProvider, IOSProvider, MockProvider
from uiautodev.remote.scrcpy import ScrcpyServer
from uiautodev.router.device import make_router
from uiautodev.router.xml import router as xml_router
from uiautodev.utils.envutils import Environment

logger = logging.getLogger(__name__)

app = FastAPI()


def enable_logger_to_console():
    _logger = logging.getLogger("uiautodev")
    _logger.setLevel(logging.DEBUG)
    _logger.addHandler(RichHandler(enable_link_path=False))


if os.getenv("UIAUTODEV_DEBUG"):
    enable_logger_to_console()

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["GET", "POST"],
    allow_headers=["*"],
)

android_router = make_router(AndroidProvider())
ios_router = make_router(IOSProvider())
harmony_router = make_router(HarmonyProvider())
mock_router = make_router(MockProvider())

app.include_router(mock_router, prefix="/api/mock", tags=["mock"])

if Environment.UIAUTODEV_MOCK:
    app.include_router(mock_router, prefix="/api/android", tags=["mock"])
    app.include_router(mock_router, prefix="/api/ios", tags=["mock"])
    app.include_router(mock_router, prefix="/api/harmony", tags=["mock"])
else:
    app.include_router(android_router, prefix="/api/android", tags=["android"])
    app.include_router(ios_router, prefix="/api/ios", tags=["ios"])
    app.include_router(harmony_router, prefix="/api/harmony", tags=["harmony"])

app.include_router(xml_router, prefix="/api/xml", tags=["xml"])


class InfoResponse(BaseModel):
    version: str
    description: str
    platform: str
    code_language: str
    cwd: str
    drivers: List[str]


@app.get("/api/info")
def info() -> InfoResponse:
    """Information about the application"""
    return InfoResponse(
        version=__version__,
        description="client for https://uiauto.dev",
        platform=platform.system(),  # Linux | Darwin | Windows
        code_language="Python",
        cwd=os.getcwd(),
        drivers=["android", "ios", "harmony"],
    )


@app.post('/api/ocr_image')
async def _ocr_image(file: UploadFile = File(...)) -> List[Node]:
    """OCR an image"""
    image_data = await file.read()
    image = convert_bytes_to_image(image_data)
    return ocr_image(image)


@app.get("/shutdown")
def shutdown() -> str:
    """Shutdown the server"""
    os.kill(os.getpid(), signal.SIGINT)
    return "Server shutting down..."


@app.get("/demo")
def demo():
    """Demo endpoint"""
    static_dir = Path(__file__).parent / "static"
    print(static_dir / "demo.html")
    return FileResponse(static_dir / "demo.html")


@app.get("/")
def index_redirect():
    """ redirect to official homepage """
    url = get_webpage_url()
    logger.debug("redirect to %s", url)
    return RedirectResponse(url)


def get_scrcpy_server(serial: str):
    # 这里主要是为了避免两次websocket建立建立，启动两个scrcpy进程
    logger.info("create scrcpy server for %s", serial)
    device = adbutils.device(serial)
    return ScrcpyServer(device)


@app.websocket("/ws/android/scrcpy/{serial}")
async def unified_ws(websocket: WebSocket, serial: str):
    """
    Args:
        serial: device serial
        websocket: WebSocket
    """
    await websocket.accept()

    try:
        logger.info(f"WebSocket serial: {serial}")

        # 获取 ScrcpyServer 实例
        server = get_scrcpy_server(serial)
        await server.handle_unified_websocket(websocket, serial)
    except WebSocketDisconnect:
        logger.info(f"WebSocket disconnected by client.")
    except Exception as e:
        logger.exception(f"WebSocket error for serial={serial}: {e}")
        await websocket.close(code=1000, reason=str(e))
    finally:
        logger.info(f"WebSocket closed for serial={serial}")


if __name__ == '__main__':
    uvicorn.run("uiautodev.app:app", port=4000, reload=True, use_colors=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Fri Mar 01 2024 14:00:10 by codeskyblue
"""

import io
import logging
from typing import Any, Dict, List

from fastapi import APIRouter, Response
from fastapi.responses import StreamingResponse
from pydantic import BaseModel

from uiautodev import command_proxy
from uiautodev.command_types import Command, CurrentAppResponse, InstallAppRequest, InstallAppResponse, TapRequest
from uiautodev.model import DeviceInfo, Node, ShellResponse
from uiautodev.provider import BaseProvider

logger = logging.getLogger(__name__)

class AndroidShellPayload(BaseModel):
    command: str


def make_router(provider: BaseProvider) -> APIRouter:
    router = APIRouter()

    @router.get("/list")
    def _list() -> List[DeviceInfo]:
        """List devices"""
        try:
            return provider.list_devices()
        except NotImplementedError as e:
            return Response(content="list_devices not implemented", media_type="text/plain", status_code=501)
        except Exception as e:
            logger.exception("list_devices failed")
            return Response(content=str(e), media_type="text/plain", status_code=500)

    @router.post("/{serial}/shell")
    def android_shell(serial: str, payload: AndroidShellPayload) -> ShellResponse:
        """Run a shell command on an Android device"""
        try:
            driver = provider.get_device_driver(serial)
            return driver.shell(payload.command)
        except NotImplementedError as e:
            return Response(content="shell not implemented", media_type="text/plain", status_code=501)
        except Exception as e:
            logger.exception("shell failed")
            return ShellResponse(output="", error=str(e))

    @router.get(
        "/{serial}/screenshot/{id}",
        responses={200: {"content": {"image/jpeg": {}}}},
        response_class=Response,
    )
    def _screenshot(serial: str, id: int) -> Response:
        """Take a screenshot of device"""
        try:
            driver = provider.get_device_driver(serial)
            pil_img = driver.screenshot(id).convert("RGB")
            buf = io.BytesIO()
            pil_img.save(buf, format="JPEG")
            image_bytes = buf.getvalue()
            return Response(content=image_bytes, media_type="image/jpeg")
        except Exception as e:
            logger.exception("screenshot failed")
            return Response(content=str(e), media_type="text/plain", status_code=500)

    @router.get("/{serial}/hierarchy")
    def dump_hierarchy(serial: str, format: str = "json") -> Node:
        """Dump the view hierarchy of an Android device"""
        try:
            driver = provider.get_device_driver(serial)
            xml_data, hierarchy = driver.dump_hierarchy()
            if format == "xml":
                return Response(content=xml_data, media_type="text/xml")
            elif format == "json":
                return hierarchy
            else:
                return Response(content=f"Invalid format: {format}", media_type="text/plain", status_code=400)
        except Exception as e:
            #logger.exception("dump_hierarchy failed")
            logger.error(f"Error dumping hierarchy: {str(e)}")
            return Response(content=str(e), media_type="text/plain", status_code=500)
    
    @router.post('/{serial}/command/tap')
    def command_tap(serial: str, params: TapRequest):
        """Run a command on the device"""
        driver = provider.get_device_driver(serial)
        command_proxy.tap(driver, params)
        return {"status": "ok"}
    
    @router.post('/{serial}/command/installApp')
    def install_app(serial: str, params: InstallAppRequest) -> InstallAppResponse:
        """Install app"""
        driver = provider.get_device_driver(serial)
        return command_proxy.app_install(driver, params)

    @router.get('/{serial}/command/currentApp')
    def current_app(serial: str) -> CurrentAppResponse:
        """Get current app"""
        driver = provider.get_device_driver(serial)
        return command_proxy.app_current(driver)

    @router.post('/{serial}/command/{command}')
    def _command_proxy_other(serial: str, command: Command, params: Dict[str, Any] = None):
        """Run a command on the device"""
        driver = provider.get_device_driver(serial)
        response = command_proxy.send_command(driver, command, params)
        return response
    
    @router.get('/{serial}/backupApp')
    def _backup_app(serial: str, packageName: str):
        """Backup app
        
        Added in 0.5.0
        """
        driver = provider.get_device_driver(serial)
        file_name = f"{packageName}.apk"
        headers = {
            'Content-Disposition': f'attachment; filename="{file_name}"'
        }
        return StreamingResponse(driver.open_app_file(packageName), headers=headers)
        


    return router



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Tue Mar 19 2024 10:43:51 by codeskyblue
"""

from __future__ import annotations

import time
import typing
from typing import Callable, Dict, List, Optional, Union

from pydantic import BaseModel

from uiautodev.command_types import AppLaunchRequest, AppTerminateRequest, By, Command, CurrentAppResponse, \
    DumpResponse, FindElementRequest, FindElementResponse, InstallAppRequest, InstallAppResponse, TapRequest, \
    WindowSizeResponse
from uiautodev.driver.base_driver import BaseDriver
from uiautodev.exceptions import ElementNotFoundError
from uiautodev.model import AppInfo, Node
from uiautodev.utils.common import node_travel

COMMANDS: Dict[Command, Callable] = {}


def register(command: Command):
    def wrapper(func):
        COMMANDS[command] = func
        return func

    return wrapper


def get_command_params_type(command: Command) -> Optional[BaseModel]:
    func = COMMANDS.get(command)
    if func is None:
        return None
    type_hints = typing.get_type_hints(func)
    return type_hints.get("params")


def send_command(driver: BaseDriver, command: Union[str, Command], params=None):
    if command not in COMMANDS:
        raise NotImplementedError(f"command {command} not implemented")
    func = COMMANDS[command]
    params_model = get_command_params_type(command)
    if params_model:
        if params is None:
            raise ValueError(f"params is required for {command}")
        if isinstance(params, dict):
            params = params_model.model_validate(params)
        elif isinstance(params, params_model):
            pass
        else:
            raise TypeError(f"params should be {params_model}", params)
    if not params:
        return func(driver)
    return func(driver, params)


@register(Command.TAP)
def tap(driver: BaseDriver, params: TapRequest):
    """Tap on the screen
    """
    x = params.x
    y = params.y
    if params.isPercent:
        wsize = driver.window_size()
        x = int(wsize[0] * params.x)
        y = int(wsize[1] * params.y)
    driver.tap(x, y)


@register(Command.APP_INSTALL)
def app_install(driver: BaseDriver, params: InstallAppRequest):
    """install app"""
    driver.app_install(params.url)
    return InstallAppResponse(success=True, id=None)


@register(Command.APP_CURRENT)
def app_current(driver: BaseDriver) -> CurrentAppResponse:
    """get current app"""
    return driver.app_current()


@register(Command.APP_LAUNCH)
def app_launch(driver: BaseDriver, params: AppLaunchRequest):
    if params.stop:
        driver.app_terminate(params.package)
    driver.app_launch(params.package)


@register(Command.APP_TERMINATE)
def app_terminate(driver: BaseDriver, params: AppTerminateRequest):
    driver.app_terminate(params.package)


@register(Command.GET_WINDOW_SIZE)
def window_size(driver: BaseDriver) -> WindowSizeResponse:
    wsize = driver.window_size()
    return WindowSizeResponse(width=wsize[0], height=wsize[1])


@register(Command.HOME)
def home(driver: BaseDriver):
    driver.home()


@register(Command.BACK)
def back(driver: BaseDriver):
    driver.back()


@register(Command.APP_SWITCH)
def app_switch(driver: BaseDriver):
    driver.app_switch()


@register(Command.VOLUME_UP)
def volume_up(driver: BaseDriver):
    driver.volume_up()


@register(Command.VOLUME_DOWN)
def volume_down(driver: BaseDriver):
    driver.volume_down()


@register(Command.VOLUME_MUTE)
def volume_mute(driver: BaseDriver):
    driver.volume_mute()


@register(Command.DUMP)
def dump(driver: BaseDriver) -> DumpResponse:
    source, _ = driver.dump_hierarchy()
    return DumpResponse(value=source)


@register(Command.WAKE_UP)
def wake_up(driver: BaseDriver):
    driver.wake_up()


def node_match(node: Node, by: By, value: str) -> bool:
    if by == By.ID:
        return node.properties.get("resource-id") == value
    if by == By.TEXT:
        return node.properties.get("text") == value
    if by == By.CLASS_NAME:
        return node.name == value
    raise ValueError(f"not support by {by!r}")


@register(Command.FIND_ELEMENTS)
def find_elements(driver: BaseDriver, params: FindElementRequest) -> FindElementResponse:
    _, root_node = driver.dump_hierarchy()
    # TODO: support By.XPATH
    nodes = []
    for node in node_travel(root_node):
        if node_match(node, params.by, params.value):
            nodes.append(node)
    return FindElementResponse(count=len(nodes), value=nodes)


@register(Command.CLICK_ELEMENT)
def click_element(driver: BaseDriver, params: FindElementRequest):
    node = None
    deadline = time.time() + params.timeout
    while time.time() < deadline:
        result = find_elements(driver, params)
        if result.value:
            node = result.value[0]
            break
        time.sleep(.5) # interval
    if not node:
        raise ElementNotFoundError(f"element not found by {params.by}={params.value}")
    center_x = (node.bounds[0] + node.bounds[2]) / 2
    center_y = (node.bounds[1] + node.bounds[3]) / 2
    tap(driver, TapRequest(x=center_x, y=center_y, isPercent=True))


@register(Command.APP_LIST)
def app_list(driver: BaseDriver) -> List[AppInfo]:
    # added in v0.5.0
    return driver.app_list()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Thu May 09 2024 11:33:17 by codeskyblue
"""


import io
import locale
import logging
from typing import List

from PIL import Image

from uiautodev.model import Node, OCRNode

logger = logging.getLogger(__name__)

def is_chinese_language() -> bool:
    language_code, _ = locale.getdefaultlocale()
    
    # Check if the language code starts with 'zh' (Chinese)
    if language_code and language_code.startswith('zh'):
        return True
    else:
        return False
    
    
def get_webpage_url() -> str:
    web_url = "https://uiauto.dev"
    if is_chinese_language():
        web_url = "https://uiauto.devsleep.com"
    return web_url


def convert_bytes_to_image(byte_data: bytes) -> Image.Image:
    return Image.open(io.BytesIO(byte_data))


def ocr_image(image: Image.Image) -> List[OCRNode]:
    # Placeholder for OCR implementation
    w, h = image.size
    try:
        from ocrmac import ocrmac
    except ImportError:
        logger.error("OCR is not supported on this platform")
        return []
    result = ocrmac.OCR(image).recognize()
    nodes = []
    for index, (text, confidence, pbounds) in enumerate(result):
        print(f"OCR result: {text}, confidence: {confidence}, bounds: {pbounds}")
        # bounds = int(pbounds[0]*w), int(pbounds[1]*h), int(pbounds[2]*w), int(pbounds[3]*h)
        nodes.append(OCRNode(key=str(index), name=text, bounds=pbounds, confidence=confidence))
    return nodes


import asyncio
import json
import logging
import os
import socket
import struct
from typing import Optional

import retry
from adbutils import AdbError, Network, adb
from adbutils._adb import AdbConnection
from adbutils._device import AdbDevice
from starlette.websockets import WebSocket, WebSocketDisconnect

from uiautodev.remote.touch_controller import ScrcpyTouchController

logger = logging.getLogger(__name__)

class ScrcpyServer:
    """
    ScrcpyServer class is responsible for managing the scrcpy server on Android devices.
    It handles the initialization, communication, and control of the scrcpy server,
    including video streaming and touch control.
    """

    def __init__(self, device: AdbDevice, scrcpy_jar_path: Optional[str] = None):
        """
        Initializes the ScrcpyServer instance.

        Args:
            scrcpy_jar_path (str, optional): Path to the scrcpy server JAR file. Defaults to None.
        """
        self.scrcpy_jar_path = scrcpy_jar_path or os.path.join(os.path.dirname(__file__), '../binaries/scrcpy_server.jar')
        self.device = device
        self.resolution_width = 0  # scrcpy 投屏转换宽度
        self.resolution_height = 0  # scrcpy 投屏转换高度
        
        self._shell_conn: AdbConnection
        self._video_conn: socket.socket
        self._control_conn: socket.socket

        self._setup_connection()
    
    def _setup_connection(self):
        self._shell_conn = self._start_scrcpy_server(control=True)
        self._video_conn = self._connect_scrcpy(self.device)
        self._control_conn = self._connect_scrcpy(self.device)
        self._parse_scrcpy_info(self._video_conn)
        self.controller = ScrcpyTouchController(self._control_conn)

    @retry.retry(exceptions=AdbError, tries=20, delay=0.1)
    def _connect_scrcpy(self, device: AdbDevice) -> socket.socket:
        return device.create_connection(Network.LOCAL_ABSTRACT, 'scrcpy')

    def _parse_scrcpy_info(self, conn: socket.socket):
        dummy_byte = conn.recv(1)
        if not dummy_byte or dummy_byte != b"\x00":
            raise ConnectionError("Did not receive Dummy Byte!")
        logger.debug('Received Dummy Byte!')
        device_name = conn.recv(64).decode('utf-8').rstrip('\x00')
        logger.debug(f'Device name: {device_name}')
        codec = conn.recv(4)
        logger.debug(f'resolution_data: {codec}')
        resolution_data = conn.recv(8)
        logger.debug(f'resolution_data: {resolution_data}')
        self.resolution_width, self.resolution_height = struct.unpack(">II", resolution_data)
        logger.debug(f'Resolution: {self.resolution_width}x{self.resolution_height}')

    def close(self):
        try:
            self._control_conn.close()
            self._video_conn.close()
            self._shell_conn.close()
        except:
            pass
    
    def __del__(self):
        self.close()

    def _start_scrcpy_server(self, control: bool = True) -> AdbConnection:
        """
        Pushes the scrcpy server JAR file to the Android device and starts the scrcpy server.

        Args:
            control (bool, optional): Whether to enable touch control. Defaults to True.

        Returns:
            AdbConnection
        """
        # 获取设备对象
        device = self.device

        # 推送 scrcpy 服务器到设备
        device.sync.push(self.scrcpy_jar_path, '/data/local/tmp/scrcpy_server.jar', check=True)
        logger.info('scrcpy server JAR pushed to device')

        # 构建启动 scrcpy 服务器的命令
        start_command = (
            'CLASSPATH=/data/local/tmp/scrcpy_server.jar '
            'app_process / '
            'com.genymobile.scrcpy.Server 2.7 '
            'log_level=info max_size=1024 max_fps=30 '
            'video_bit_rate=8000000 tunnel_forward=true '
            'send_frame_meta=false '
            f'control={"true" if control else "false"} '
            'audio=false show_touches=false stay_awake=false '
            'power_off_on_close=false clipboard_autosync=false'
        )
        conn = device.shell(start_command, stream=True)
        logger.debug("scrcpy output: %s", conn.conn.recv(100))
        return conn # type: ignore

    async def handle_unified_websocket(self, websocket: WebSocket, serial=''):
        logger.info(f"[Unified] WebSocket connection from {websocket} for serial: {serial}")

        video_task = asyncio.create_task(self._stream_video_to_websocket(self._video_conn, websocket))
        control_task  = asyncio.create_task(self._handle_control_websocket(websocket))
        
        try:
            # 不使用 return_exceptions=True，让异常能够正确传播
            await asyncio.gather(video_task, control_task)
        finally:
            # 取消任务
            for task in (video_task, control_task):
                if not task.done():
                    task.cancel()
            logger.info(f"[Unified] WebSocket closed for serial={serial}")

    async def _stream_video_to_websocket(self, conn: socket.socket, ws: WebSocket):
        # Set socket to non-blocking mode
        conn.setblocking(False)
        
        while True:
            # check if ws closed
            if ws.client_state.name != "CONNECTED":
                logger.info('WebSocket no longer connected. Exiting video stream.')
                break
            # Use asyncio to read data asynchronously
            data = await asyncio.get_event_loop().sock_recv(conn, 1024*1024)
            if not data:
                logger.warning('No data received, connection may be closed.')
                raise ConnectionError("Video stream ended unexpectedly")
            # send data to ws
            await ws.send_bytes(data)

    async def _handle_control_websocket(self, ws: WebSocket):
        while True:
            try:
                message = await ws.receive_text()
                logger.debug(f"[Unified] Received message: {message}")
                message = json.loads(message)

                width, height = self.resolution_width, self.resolution_height
                message_type = message.get('type')
                if message_type == 'touchMove':
                    xP = message['xP']
                    yP = message['yP']
                    self.controller.move(int(xP * width), int(yP * height), width, height)
                elif message_type == 'touchDown':
                    xP = message['xP']
                    yP = message['yP']
                    self.controller.down(int(xP * width), int(yP * height), width, height)
                elif message_type == 'touchUp':
                    xP = message['xP']
                    yP = message['yP']
                    self.controller.up(int(xP * width), int(yP * height), width, height)
                elif message_type == 'keyEvent':
                    event_number = message['data']['eventNumber']
                    self.device.shell(f'input keyevent {event_number}')
                elif message_type == 'text':
                    text = message['detail']
                    self.device.shell(f'am broadcast -a SONIC_KEYBOARD --es msg \'{text}\'')
                elif message_type == 'ping':
                    await ws.send_text(json.dumps({"type": "pong"}))
            except json.JSONDecodeError as e:
                logger.error(f"Invalid JSON message: {e}")
                continue


from enum import IntEnum


class KeyCode(IntEnum):
    """Android key codes ported from Android's KeyEvent class
    
    This enum contains all the key codes defined in Android's KeyEvent class,
    which are used for sending key events to Android devices through scrcpy.
    
    The comments for each key code are taken directly from the Android source code
    to maintain compatibility and provide accurate descriptions.
    """
    # Unknown key code
    UNKNOWN = 0
    # Soft Left key - Usually situated below the display on phones
    SOFT_LEFT = 1
    # Soft Right key - Usually situated below the display on phones
    SOFT_RIGHT = 2
    # Home key - This key is handled by the framework and is never delivered to applications
    HOME = 3
    # Back key
    BACK = 4
    # Call key
    CALL = 5
    # End Call key
    ENDCALL = 6
    # '0' key
    KEY_0 = 7
    # '1' key
    KEY_1 = 8
    # '2' key
    KEY_2 = 9
    # '3' key
    KEY_3 = 10
    # '4' key
    KEY_4 = 11
    # '5' key
    KEY_5 = 12
    # '6' key
    KEY_6 = 13
    # '7' key
    KEY_7 = 14
    # '8' key
    KEY_8 = 15
    # '9' key
    KEY_9 = 16
    # '*' key
    STAR = 17
    # '#' key
    POUND = 18
    # Directional Pad Up key - May also be synthesized from trackball motions
    DPAD_UP = 19
    # Directional Pad Down key - May also be synthesized from trackball motions
    DPAD_DOWN = 20
    # Directional Pad Left key - May also be synthesized from trackball motions
    DPAD_LEFT = 21
    # Directional Pad Right key - May also be synthesized from trackball motions
    DPAD_RIGHT = 22
    # Directional Pad Center key - May also be synthesized from trackball motions
    DPAD_CENTER = 23
    # Volume Up key - Adjusts the speaker volume up
    VOLUME_UP = 24
    # Volume Down key - Adjusts the speaker volume down
    VOLUME_DOWN = 25
    # Power key
    POWER = 26
    # Camera key - Used to launch a camera application or take pictures
    CAMERA = 27
    # Clear key
    CLEAR = 28
    A = 29
    B = 30
    C = 31
    D = 32
    E = 33
    F = 34
    G = 35
    H = 36
    I = 37
    J = 38
    K = 39
    L = 40
    M = 41
    N = 42
    O = 43
    P = 44
    Q = 45
    R = 46
    S = 47
    T = 48
    U = 49
    V = 50
    W = 51
    X = 52
    Y = 53
    Z = 54
    COMMA = 55
    PERIOD = 56
    ALT_LEFT = 57
    ALT_RIGHT = 58
    SHIFT_LEFT = 59
    SHIFT_RIGHT = 60
    TAB = 61
    SPACE = 62
    SYM = 63
    EXPLORER = 64
    ENVELOPE = 65
    # Enter key
    ENTER = 66
    # Backspace key - Deletes characters before the insertion point
    DEL = 67
    GRAVE = 68
    MINUS = 69
    EQUALS = 70
    LEFT_BRACKET = 71
    RIGHT_BRACKET = 72
    BACKSLASH = 73
    SEMICOLON = 74
    APOSTROPHE = 75
    SLASH = 76
    AT = 77
    NUM = 78
    HEADSETHOOK = 79
    FOCUS = 80
    PLUS = 81
    # Menu key
    MENU = 82
    NOTIFICATION = 83
    SEARCH = 84
    MEDIA_PLAY_PAUSE = 85
    MEDIA_STOP = 86
    MEDIA_NEXT = 87
    MEDIA_PREVIOUS = 88
    MEDIA_REWIND = 89
    MEDIA_FAST_FORWARD = 90
    MUTE = 91
    PAGE_UP = 92
    PAGE_DOWN = 93
    PICTSYMBOLS = 94
    SWITCH_CHARSET = 95
    BUTTON_A = 96
    BUTTON_B = 97
    BUTTON_C = 98
    BUTTON_X = 99
    BUTTON_Y = 100
    BUTTON_Z = 101
    BUTTON_L1 = 102
    BUTTON_R1 = 103
    BUTTON_L2 = 104
    BUTTON_R2 = 105
    BUTTON_THUMBL = 106
    BUTTON_THUMBR = 107
    BUTTON_START = 108
    BUTTON_SELECT = 109
    BUTTON_MODE = 110
    ESCAPE = 111
    FORWARD_DEL = 112
    CTRL_LEFT = 113
    CTRL_RIGHT = 114
    CAPS_LOCK = 115
    SCROLL_LOCK = 116
    META_LEFT = 117
    META_RIGHT = 118
    FUNCTION = 119
    SYSRQ = 120
    BREAK = 121
    MOVE_HOME = 122
    MOVE_END = 123
    INSERT = 124
    FORWARD = 125
    MEDIA_PLAY = 126
    MEDIA_PAUSE = 127
    MEDIA_CLOSE = 128
    MEDIA_EJECT = 129
    MEDIA_RECORD = 130
    F1 = 131
    F2 = 132
    F3 = 133
    F4 = 134
    F5 = 135
    F6 = 136
    F7 = 137
    F8 = 138
    F9 = 139
    F10 = 140
    F11 = 141
    F12 = 142
    NUM_LOCK = 143
    NUMPAD_0 = 144
    NUMPAD_1 = 145
    NUMPAD_2 = 146
    NUMPAD_3 = 147
    NUMPAD_4 = 148
    NUMPAD_5 = 149
    NUMPAD_6 = 150
    NUMPAD_7 = 151
    NUMPAD_8 = 152
    NUMPAD_9 = 153
    NUMPAD_DIVIDE = 154
    NUMPAD_MULTIPLY = 155
    NUMPAD_SUBTRACT = 156
    NUMPAD_ADD = 157
    NUMPAD_DOT = 158
    NUMPAD_COMMA = 159
    NUMPAD_ENTER = 160
    NUMPAD_EQUALS = 161
    NUMPAD_LEFT_PAREN = 162
    NUMPAD_RIGHT_PAREN = 163
    VOLUME_MUTE = 164
    INFO = 165
    CHANNEL_UP = 166
    CHANNEL_DOWN = 167
    ZOOM_IN = 168
    ZOOM_OUT = 169
    TV = 170
    WINDOW = 171
    GUIDE = 172
    DVR = 173
    BOOKMARK = 174
    CAPTIONS = 175
    SETTINGS = 176
    TV_POWER = 177
    TV_INPUT = 178
    STB_POWER = 179
    STB_INPUT = 180
    AVR_POWER = 181
    AVR_INPUT = 182
    PROG_RED = 183
    PROG_GREEN = 184
    PROG_YELLOW = 185
    PROG_BLUE = 186
    APP_SWITCH = 187
    BUTTON_1 = 188
    BUTTON_2 = 189
    BUTTON_3 = 190
    BUTTON_4 = 191
    BUTTON_5 = 192
    BUTTON_6 = 193
    BUTTON_7 = 194
    BUTTON_8 = 195
    BUTTON_9 = 196
    BUTTON_10 = 197
    BUTTON_11 = 198
    BUTTON_12 = 199
    BUTTON_13 = 200
    BUTTON_14 = 201
    BUTTON_15 = 202
    BUTTON_16 = 203
    LANGUAGE_SWITCH = 204
    MANNER_MODE = 205
    MODE_3D = 206
    CONTACTS = 207
    CALENDAR = 208
    MUSIC = 209
    CALCULATOR = 210
    ZENKAKU_HANKAKU = 211
    EISU = 212
    MUHENKAN = 213
    HENKAN = 214
    KATAKANA_HIRAGANA = 215
    YEN = 216
    RO = 217
    KANA = 218
    ASSIST = 219
    BRIGHTNESS_DOWN = 220
    BRIGHTNESS_UP = 221
    MEDIA_AUDIO_TRACK = 222
    SLEEP = 223
    WAKEUP = 224
    PAIRING = 225
    MEDIA_TOP_MENU = 226
    KEY_11 = 227
    KEY_12 = 228
    LAST_CHANNEL = 229
    TV_DATA_SERVICE = 230
    VOICE_ASSIST = 231
    TV_RADIO_SERVICE = 232
    TV_TELETEXT = 233
    TV_NUMBER_ENTRY = 234
    TV_TERRESTRIAL_ANALOG = 235
    TV_TERRESTRIAL_DIGITAL = 236
    TV_SATELLITE = 237
    TV_SATELLITE_BS = 238
    TV_SATELLITE_CS = 239
    TV_SATELLITE_SERVICE = 240
    TV_NETWORK = 241
    TV_ANTENNA_CABLE = 242
    TV_INPUT_HDMI_1 = 243
    TV_INPUT_HDMI_2 = 244
    TV_INPUT_HDMI_3 = 245
    TV_INPUT_HDMI_4 = 246
    TV_INPUT_COMPOSITE_1 = 247
    TV_INPUT_COMPOSITE_2 = 248
    TV_INPUT_COMPONENT_1 = 249
    TV_INPUT_COMPONENT_2 = 250
    TV_INPUT_VGA_1 = 251
    TV_AUDIO_DESCRIPTION = 252
    TV_AUDIO_DESCRIPTION_MIX_UP = 253
    TV_AUDIO_DESCRIPTION_MIX_DOWN = 254
    TV_ZOOM_MODE = 255
    TV_CONTENTS_MENU = 256
    TV_MEDIA_CONTEXT_MENU = 257
    TV_TIMER_PROGRAMMING = 258
    HELP = 259
    NAVIGATE_PREVIOUS = 260
    NAVIGATE_NEXT = 261
    NAVIGATE_IN = 262
    NAVIGATE_OUT = 263
    STEM_PRIMARY = 264
    STEM_1 = 265
    STEM_2 = 266
    STEM_3 = 267
    DPAD_UP_LEFT = 268
    DPAD_DOWN_LEFT = 269
    DPAD_UP_RIGHT = 270
    DPAD_DOWN_RIGHT = 271
    MEDIA_SKIP_FORWARD = 272
    MEDIA_SKIP_BACKWARD = 273
    MEDIA_STEP_FORWARD = 274
    MEDIA_STEP_BACKWARD = 275
    SOFT_SLEEP = 276
    CUT = 277
    COPY = 278
    PASTE = 279
    SYSTEM_NAVIGATION_UP = 280
    SYSTEM_NAVIGATION_DOWN = 281
    SYSTEM_NAVIGATION_LEFT = 282
    SYSTEM_NAVIGATION_RIGHT = 283
    ALL_APPS = 284
    
    # =========================================================================
    # Aliases for original Android KeyEvent names
    # =========================================================================
    # These aliases are provided to maintain compatibility with the original
    # Android KeyEvent naming convention (AKEYCODE_*). This makes it easier
    # to reference keys using the same names as in Android documentation.
    
    # Numeric key aliases
    KEYCODE_0 = KEY_0         # '0' key
    KEYCODE_1 = KEY_1         # '1' key
    KEYCODE_2 = KEY_2         # '2' key
    KEYCODE_3 = KEY_3         # '3' key
    KEYCODE_4 = KEY_4         # '4' key
    KEYCODE_5 = KEY_5         # '5' key
    KEYCODE_6 = KEY_6         # '6' key
    KEYCODE_7 = KEY_7         # '7' key
    KEYCODE_8 = KEY_8         # '8' key
    KEYCODE_9 = KEY_9         # '9' key
    KEYCODE_11 = KEY_11       # '11' key
    KEYCODE_12 = KEY_12       # '12' key


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Tue Mar 19 2024 10:53:03 by codeskyblue
"""

from __future__ import annotations

import logging
import os
import platform
import subprocess
import sys
import threading
import time
from pprint import pprint

import click
import httpx
import pydantic
import uvicorn

from uiautodev import __version__, command_proxy
from uiautodev.command_types import Command
from uiautodev.common import get_webpage_url
from uiautodev.provider import AndroidProvider, BaseProvider, IOSProvider
from uiautodev.utils.common import convert_params_to_model, print_json

logger = logging.getLogger(__name__)

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option("--verbose", "-v", is_flag=True, default=False, help="verbose mode")
def cli(verbose: bool):
    if verbose:
        os.environ['UIAUTODEV_DEBUG'] = '1'
        logger.debug("Verbose mode enabled")


def run_driver_command(provider: BaseProvider, command: Command, params: list[str] = None):
    if command == Command.LIST:
        devices = provider.list_devices()
        print("==> Devices <==")
        pprint(devices)
        return
    driver = provider.get_single_device_driver()
    params_obj = None
    model = command_proxy.get_command_params_type(command)
    if model:
        if not params:
            print(f"params is required for {command}")
            pprint(model.model_json_schema())
            return
        params_obj = convert_params_to_model(params, model)

    try:
        print("Command:", command.value)
        print("Params ↓")
        print_json(params_obj)
        result = command_proxy.send_command(driver, command, params_obj)
        print("Result ↓")
        print_json(result)
    except pydantic.ValidationError as e:
        print(f"params error: {e}")
        print(f"\n--- params should be match schema ---")
        pprint(model.model_json_schema()["properties"])


@cli.command(help="COMMAND: " + ", ".join(c.value for c in Command))
@click.argument("command", type=Command, required=True)
@click.argument("params", required=False, nargs=-1)
def android(command: Command, params: list[str] = None):
    provider = AndroidProvider()
    run_driver_command(provider, command, params)


@cli.command(help="COMMAND: " + ", ".join(c.value for c in Command))
@click.argument("command", type=Command, required=True)
@click.argument("params", required=False, nargs=-1)
def ios(command: Command, params: list[str] = None):
    provider = IOSProvider()
    run_driver_command(provider, command, params)


@cli.command(help="run case (beta)")
def case():
    from uiautodev.case import run
    run()


@cli.command(help="COMMAND: " + ", ".join(c.value for c in Command))
@click.argument("command", type=Command, required=True)
@click.argument("params", required=False, nargs=-1)
def appium(command: Command, params: list[str] = None):
    from uiautodev.driver.appium import AppiumProvider
    from uiautodev.exceptions import AppiumDriverException

    provider = AppiumProvider()
    try:
        run_driver_command(provider, command, params)
    except AppiumDriverException as e:
        print(f"Error: {e}")


@cli.command('version')
def print_version():
    """ Print version """
    print(__version__)


@cli.command('self-update')
def self_update():
    """ Update uiautodev to latest version """
    subprocess.run([sys.executable, '-m', "pip", "install", "--upgrade", "uiautodev"])


@cli.command(help="start uiauto.dev local server [Default]")
@click.option("--port", default=20242, help="port number", show_default=True)
@click.option("--host", default="127.0.0.1", help="host", show_default=True)
@click.option("--reload", is_flag=True, default=False, help="auto reload, dev only")
@click.option("-f", "--force", is_flag=True, default=False, help="shutdown alrealy runningserver")
@click.option("-s", "--no-browser", is_flag=True, default=False, help="silent mode, do not open browser")
def server(port: int, host: str, reload: bool, force: bool, no_browser: bool):
    print("uiautodev version:", __version__)
    if force:
        try:
            httpx.get(f"http://{host}:{port}/shutdown", timeout=3)
        except httpx.HTTPError:
            pass

    use_color = True
    if platform.system() == 'Windows':
        use_color = False

    if not no_browser:
        th = threading.Thread(target=open_browser_when_server_start, args=(f"http://{host}:{port}",))
        th.daemon = True
        th.start()
    uvicorn.run("uiautodev.app:app", host=host, port=port, reload=reload, use_colors=use_color)


def open_browser_when_server_start(server_url: str):
    deadline = time.time() + 10
    while time.time() < deadline:
        try:
            httpx.get(f"{server_url}/api/info", timeout=1)
            break
        except Exception as e:
            time.sleep(0.5)
    import webbrowser
    web_url = get_webpage_url()
    logger.info("open browser: %s", web_url)
    webbrowser.open(web_url)


def main():
    # set logger level to INFO
    # logging.basicConfig(level=logging.INFO)
    logger.setLevel(logging.INFO)

    has_command = False
    for name in sys.argv[1:]:
        if not name.startswith("-"):
            has_command = True

    if not has_command:
        cli.main(args=sys.argv[1:] + ["server"], prog_name="uiauto.dev")
    else:
        cli()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Tue Mar 19 2024 15:51:59 by codeskyblue
"""

from __future__ import annotations

import io
import json
import logging
from pprint import pprint
from typing import Tuple

import httpretty
import httpx
from appium import webdriver
from appium.options.android import UiAutomator2Options
from appium.options.ios import XCUITestOptions
from appium.webdriver.common.appiumby import AppiumBy as By
from PIL import Image
from selenium.webdriver.common.proxy import Proxy, ProxyType

from uiautodev.command_types import CurrentAppResponse
from uiautodev.driver.android import parse_xml
from uiautodev.driver.base_driver import BaseDriver
from uiautodev.exceptions import AppiumDriverException
from uiautodev.model import DeviceInfo, Node, ShellResponse, WindowSize
from uiautodev.provider import BaseProvider

logger = logging.getLogger(__name__)

class AppiumProvider(BaseProvider):
    sessions = []

    def __init__(self, command_executor: str = "http://localhost:4723/wd/hub"):
        # command_executor = "http://localhost:4700"
        # command_executor = "http://localhost:4720/wd/hub"
        self.command_executor = command_executor.rstrip('/')
        self.sessions.clear()

    def list_devices(self) -> list[DeviceInfo]:
        """ appium just return all session_ids """
        response = httpx.get(f"{self.command_executor}/sessions", verify=False)
        if response.status_code >= 400:
            raise AppiumDriverException(f"Failed request to appium server: {self.command_executor} status: {response.status_code}")
        ret = []
        self.sessions = response.json()['value']
        for item in self.sessions:
            item['sessionId'] = item.pop('id')
            print("Active sessionId", item['sessionId'])
            serial = item['capabilities']['platformName'] + ':' + item['sessionId']
            ret.append(DeviceInfo(
                serial=serial,
                model=item['capabilities']['deviceModel'],
                name=item['capabilities']['deviceName'],
            ))
        return ret
    
    def get_device_driver(self, serial: str, session_id: str = None) -> BaseDriver:
        """ TODO: attach to the existing session """
        platform_name, session_id = serial.split(':', 1)
        filtered_sessions = [session for session in self.sessions if session['sessionId'] == session_id]
        if len(filtered_sessions) == 1:
            session = filtered_sessions[0]
            driver = self.attach_session(session)
            return AppiumDriver(driver, is_attached=True)
        else:
            options = UiAutomator2Options() if platform_name == "Android" else XCUITestOptions()
            driver = webdriver.Remote(self.command_executor, options=options)
            return AppiumDriver(driver)
        
    @httpretty.activate(allow_net_connect=False)
    def attach_session(self, session: dict) -> webdriver.Remote:
        """
        https://github.com/appium/python-client/issues/212
        the author say it can't
        """
        body = json.dumps({'value': session}, indent=4)
        logger.debug("Mock response: POST /wd/hub/session", body)
        httpretty.register_uri(httpretty.POST,
                               self.command_executor + '/session',
                               body=body,
                               headers={'Content-Type': 'application/json'})
        options = UiAutomator2Options()# if platform_name == "Android" else XCUITestOptions()
        driver = webdriver.Remote(command_executor=self.command_executor, strict_ssl=False, options=options)
        return driver

    def get_single_device_driver(self) -> BaseDriver:
        devices = self.list_devices()
        if len(devices) == 0:
            return self.get_device_driver("Android:12345")
        #     raise AppiumDriverException("No device found")
        return self.get_device_driver(devices[0].serial)


class AppiumDriver(BaseDriver):
    def __init__(self, driver: webdriver.Remote, is_attached: bool = False):
        self.driver = driver
        self.is_attached = is_attached
    
    # def __del__(self):
    #     if not self.is_attached:
    #         self.driver.quit()

    def screenshot(self, id: int) -> Image:
        png_data = self.driver.get_screenshot_as_png()
        return Image.open(io.BytesIO(png_data))

    def window_size(self) -> WindowSize:
        size = self.driver.get_window_size()
        return WindowSize(width=size["width"], height=size["height"])
        
    def dump_hierarchy(self) -> Tuple[str, Node]:
        source = self.driver.page_source
        wsize = self.window_size()
        return source, parse_xml(source, wsize)
    
    def shell(self, command: str) -> ShellResponse:
        # self.driver.execute_script(command)
        raise NotImplementedError()

    def tap(self, x: int, y: int):
        self.driver.tap([(x, y)], 100)
        print("Finished")
    
    def app_install(self, app_path: str):
        self.driver.install_app(app_path)
    
    def app_current(self) -> CurrentAppResponse:
        package = self.driver.current_package
        activity = self.driver.current_activity
        return CurrentAppResponse(package=package, activity=activity)

    def home(self):
        self.driver.press_keycode(3)